"""
Core base classes of the turboPy framework

Notes
-----
A preprint of the paper describing this code is available on
the arxiv [1]_.

References
----------
.. [1] A. S. Richardson et al., "TurboPy: A Lightweight Python Framework
       for Computational Physics." Preprint available online:
       https://arxiv.org/abs/2002.08842

"""
from pathlib import Path
from abc import ABC, abstractmethod
import numpy as np


class Simulation:
    """Main turboPy simulation class

    This Class "owns" all the physics modules, compute tools, and
    diagnostics. It also coordinates them. The main simulation loop is
    driven by an instance of this class.

    Parameters
    ----------
    input_data : `dict`
        This dictionary contains all parameters needed to set up a
        turboPy simulation. Each key describes a section, and the value
        is another dictionary with the needed parameters for that
        section.

        Expected keys are:

        ``"Grid"``
            Dictionary containg parameters needed to defined the grid.
            Currently only 1D grids are defined in turboPy.

            The expected parameters are:

            - ``"N"`` | {``"dr"`` | ``"dx"``} :
                The number of grid points (`int`) | the grid spacing
                (`float`)
            - ``"min"`` | ``"x_min"`` | ``"r_min"`` :
                The coordinate value of the minimum grid point (`float`)
            - ``"max"`` | ``"x_max"`` | ``"r_max"`` :
                The coordinate value of the maximum grid point (`float`)

        ``"Clock"``
            Dictionary of parameters needed to define the simulation
            clock.

            The expected parameters are:

            - ``"start_time"`` :
                The time for the start of the simulation (`float`)
            - ``"end_time"`` :
                The time for the end of the simulation (`float`)
            - ``"num_steps"`` | ``"dt"`` :
                The number of time steps (`int`) | the size of the time
                step (`float`)
            - ``"print_time"`` :
                `bool`, optional, default is ``False``

        ``"PhysicsModules"`` : `dict` [`str`, `dict`]
            Dictionary of :class:`PhysicsModule` items needed for the
            simulation.

            Each key in the dictionary should map to a
            :class:`PhysicsModule` subclass key in the
            :class:`PhysicsModule` registry.

            The value is a dictionary of parameters which is passed to
            the constructor for the :class:`PhysicsModule`.

        ``"Diagnostics"`` : `dict` [`str`, `dict`], optional
            Dictionary of :class:`Diagnostic` items needed for the
            simulaiton.

            Each key in the dictionary should map to a
            :class:`Diagnostic` subclass key in the :class:`Diagnostic`
            registry.

            The value is a dictionary of parameters which is passed to
            the constructor for the :class:`Diagnostic`.

            If the key is not found in the registry, then the key/value
            pair is interpreted as a default parameter value, and is
            added to dictionary of parameters for all of the
            :class:`Diagnostic` constructors.

        ``"Tools"`` : `dict` [`str`, `dict`], optional
            Dictionary of :class:`ComputeTool` items needed for the
            simulation.

            Each key in the dictionary should map to a
            :class:`ComputeTool` subclass key in the
            :class:`ComputeTool` registry.

            The value is a dictionary of parameters which is passed to
            the constructor for the :class:`ComputeTool`.

    Attributes
    ----------
    physics_modules : list of :class:`PhysicsModule` subclass objects
        A list of :class:`PhysicsModule` objects for this simulation.
    diagnostics : list of :class:`Diagnostic` subclass objects
        A list of :class:`Diagnostic` objects for this simulation.
    compute_tools : list of :class:`ComputeTool` subclass objects
        A list of :class:`ComputeTool` objects for this simulation.
    """

    def __init__(self, input_data: dict):
        self.physics_modules = []
        self.compute_tools = []
        self.diagnostics = []

        self.grid = None
        self.clock = None
        self.units = None

        self.input_data = input_data

    def run(self):
        """
        Runs the simulation

        This initializes the simulation, runs the main loop, and then
        finalizes the simulation.
        """
        print("Simulation is initializing")
        self.prepare_simulation()
        print("Initialization complete")

        print("Simulation is started")
        while self.clock.is_running():
            self.fundamental_cycle()

        self.finalize_simulation()
        print("Simulation complete")

    def fundamental_cycle(self):
        """
        Perform one step of the main time loop

        Executes each diagnostic and physics module, and advances
        the clock.
        """
        for d in self.diagnostics:
            d.diagnose()
        for m in self.physics_modules:
            m.reset()
        for m in self.physics_modules:
            m.update()
        self.clock.advance()

    def prepare_simulation(self):
        """
        Prepares the simulation by reading the input and initializing
        physics modules and diagnostics.
        """
        print("Reading Grid...")
        self.read_grid_from_input()

        print("Reading Tools...")
        self.read_tools_from_input()

        print("Reading PhysicsModules...")
        self.read_modules_from_input()

        print("Reading Diagnostics...")
        self.read_diagnostics_from_input()

        print("Initializing Simulation Clock...")
        self.read_clock_from_input()

        print("Initializing Tools...")
        for t in self.compute_tools:
            t.initialize()

        print("Initializing PhysicsModules...")
        for m in self.physics_modules:
            m.exchange_resources()
        for m in self.physics_modules:
            m.initialize()

        print("Initializing Diagnostics...")
        for d in self.diagnostics:
            d.initialize()

    def finalize_simulation(self):
        """
        Close out the simulation

        Runs the :class:`Diagnostic.finalize()` method for each
        diagnostic.
        """
        for d in self.diagnostics:
            d.finalize()

    def read_grid_from_input(self):
        """Construct the grid based on input parameters"""
        self.grid = Grid(self.input_data["Grid"])

    def read_clock_from_input(self):
        """Construct the clock based on input parameters"""
        self.clock = SimulationClock(self, self.input_data["Clock"])

    def read_tools_from_input(self):
        """Construct :class:`ComputeTools` based on input"""
        if "Tools" in self.input_data:
            for tool_name, params in self.input_data["Tools"].items():
                tool_class = ComputeTool.lookup(tool_name)
                params["type"] = tool_name
                # TODO: somehow make tool names unique, or prevent
                #  more than one each
                self.compute_tools.append(tool_class(
                    owner=self, input_data=params))

    def read_modules_from_input(self):
        """Construct :class:`PhysicsModule` instances based on input"""
        for physics_module_name, physics_module_data in \
                self.input_data["PhysicsModules"].items():
            physics_module_class = PhysicsModule.lookup(
                                            physics_module_name)
            physics_module_data["name"] = physics_module_name
            self.physics_modules.append(physics_module_class(
                            owner=self, input_data=physics_module_data))
        self.sort_modules()

    def read_diagnostics_from_input(self):
        """Construct :class:`Diagnostic` instances based on input"""
        if "Diagnostics" in self.input_data:
            # This dictionary has two types of keys:
            #    keys that are valid diagnostic types
            #    other keys, which should be passed along
            #    as "default" parameters
            diags = {k: v for k, v in
                     self.input_data["Diagnostics"].items()
                     if Diagnostic.is_valid_name(k)}
            params = {k: v for k, v in
                      self.input_data["Diagnostics"].items()
                      if not Diagnostic.is_valid_name(k)}

            # todo: implement a system for making default file names
            if "directory" in params:
                d = Path(params["directory"])
                d.mkdir(parents=True, exist_ok=True)

            for diag_type, d in diags.items():
                diagnostic_class = Diagnostic.lookup(diag_type)
                if not type(d) is list:
                    d = [d]
                for di in d:
                    # Values in di supersede values in params because
                    # of the order in which these are combined
                    di = {**params, **di, "type": diag_type}
                    if "directory" in di and "filename" in di:
                        di["filename"] = str(Path(di["directory"])
                                             / Path(di["filename"]))
                    self.diagnostics.append(
                        diagnostic_class(owner=self, input_data=di))

    def sort_modules(self):
        """Sort :class:`Simulation.physics_modules` by some logic

        Unused stub for future implementation"""
        pass

    def find_tool_by_name(self, tool_name: str):
        """Returns the :class:`ComputeTool` associated with the
        given name"""
        tools = [t for t in self.compute_tools if t.name == tool_name]
        if len(tools) == 1:
            return tools[0]
        return None


class DynamicFactory(ABC):
    """Abstract class which provides dynamic factory functionality

    This base class provides a dynamic factory pattern functionality to
    classes that derive from this.
    """
    @property
    @abstractmethod
    def _factory_type_name(self):
        """Override this in derived classes with a string that
        describes type of the derived factory"""
        pass

    @property
    @abstractmethod
    def _registry(self):
        """Override this in derived classes with a dictionary that
        holds references to derived subclasses"""
        pass

    @classmethod
    def register(cls, name_to_register: str, class_to_register,
                 override=False):
        """Add a derived class to the registry"""
        if name_to_register in cls._registry and not override:
            raise ValueError("{0} '{1}' already registered".format(
                cls._factory_type_name, name_to_register))
        if not issubclass(class_to_register, cls):
            raise TypeError("{0} is not a subclass of {1}".format(
                class_to_register, cls))
        cls._registry[name_to_register] = class_to_register

    @classmethod
    def lookup(cls, name: str):
        """Look up a name in the registry, and return the associated
        derived class"""
        try:
            return cls._registry[name]
        except KeyError:
            raise KeyError("{0} '{1}' not found in registry".format(
                cls._factory_type_name, name))

    @classmethod
    def is_valid_name(cls, name: str):
        """Check if the name is in the registry"""
        return name in cls._registry


class PhysicsModule(DynamicFactory):
    """
    This is the base class for all physics modules
    Based on Module class in TurboWAVE
    Because python mutable/immutable is different than C++ pointers, the
    implementation here is different. Here, a "resource" is a
    dictionary, and can have more than one thing being shared. Note that
    the value stored in the dictionary needs to be mutable. Make sure
    not to reinitialize it, because other physics modules will be
    holding a reference to it.

    Parameters
    ----------
    _registery : dict
        Registered derived ComputeTool classes.
    _factory_type_name : str
        Type of PhysicsModule child class
    owner : :class:`Simulation`
        Simulation class that PhysicsModule belongs to.
    input_data : dict
       Dictionary that contains user defined parameters about this
       object such as its name.

    Attributes
    ----------
    owner : :class:`Simulation`
        Simulation class that PhysicsModule belongs to.
    module_type : None
        Module type.
    input_data : dict
       Dictionary that contains user defined parameters about this
       object such as its name.
    """
    _factory_type_name = "Physics Module"
    _registry = {}

    def __init__(self, owner: Simulation, input_data: dict):
        self.owner = owner
        self.module_type = None
        self.input_data = input_data

    def publish_resource(self, resource: dict):
        """
        Method which implements the details of sharing resources

        Parameters
        ----------
        resource : dict
            resource dictionary to be shared
        """
        for physics_module in self.owner.physics_modules:
            physics_module.inspect_resource(resource)
        for diagnostic in self.owner.diagnostics:
            diagnostic.inspect_resource(resource)

    def inspect_resource(self, resource: dict):
        """Method for accepting resources shared by other PhysicsModules
        If your subclass needs the data described by the key, now's
        their chance to save a pointer to the data.

        Parameters
        ----------
        resource : dict
            resource dictionary to be shared
        """
        pass

    def exchange_resources(self):
        """Main method for sharing resources with other PhysicsModules
        This is the function where you call publish_resource, to tell
        other physics modules about data you want to share.
        """
        pass

    def update(self):
        """Do the main work of the PhysicsModule
        This is called at every time step in the main loop.
        """
        raise NotImplementedError

    def reset(self):
        """Perform any needed reset operations
        This is called at every time step in the main loop, before any
        of the calls to `update`.
        """
        pass

    def initialize(self):
        """Perform initialization operations for this PhysicsModule
        This is called before the main simulation loop
        """
        pass


class ComputeTool(DynamicFactory):
    """This is the base class for compute tools

    These are the compute-heavy functions, which have implementations
    of numerical methods which can be shared between physics modules.

    Parameters
    ----------
    owner : :class:`Simulation`
        Simulation class that ComputeTool belongs to.
    input_data : dict
        Dictionary that contains user defined parameters about this
        object such as its name.

    Attributes
    ----------
    _registery : dict
        Registered derived ComputeTool classes.
    _factory_type_name : str
        Type of ComputeTool child class
    owner : :class:`Simulation`
        Simulation class that ComputeTool belongs to.
    input_data : dict
        Dictionary that contains user defined parameters about this
        object such as its name.
    name : str
        Type of ComputeTool.
    """

    _factory_type_name = "Compute Tool"
    _registry = {}

    def __init__(self, owner: Simulation, input_data: dict):
        self.owner = owner
        self.input_data = input_data
        self.name = input_data["type"]

    def initialize(self):
        """Perform any initialization operations needed for this tool"""
        pass


class SimulationClock:
    """
    Clock class for turboPy

    Parameters
    ----------
    owner : :class:`Simulation`
        Simulation class that SimulationClock belongs to.
    clock_data : dict
        Dictionary of parameters needed to define the simulation
        clock.

        The expected parameters are:

        - ``"start_time"`` :
            The time for the start of the simulation (`float`)
        - ``"end_time"`` :
            The time for the end of the simulation (`float`)
        - ``"num_steps"`` | ``"dt"`` :
            The number of time steps (`int`) | the size of the time
            step (`float`)
        - ``"print_time"`` :
            `bool`, optional, default is ``False``

    Attributes
    ----------
    owner : :class:`Simulation`
        Simulation class that SimulationClock belongs to.
    start_time : float
        Clock start time.
    time : float
        Current time on clock.
    end_time : float
        Clock end time.
    this_step : int
        Current time step since start.
    print_time : bool
        If True will print current time after each increment.
    num_steps : int
        Number of steps clock will take in the interval.
    dt : float
        Time passed at each increment.
    """

    def __init__(self, owner: Simulation, clock_data: dict):
        self.owner = owner
        self.start_time = clock_data["start_time"]
        self.time = self.start_time
        self.end_time = clock_data["end_time"]
        self.this_step = 0
        self.print_time = False
        if "print_time" in clock_data:
            self.print_time = clock_data["print_time"]

        if "num_steps" in clock_data:
            self.num_steps = clock_data["num_steps"]
            self.dt = (
                    (clock_data["end_time"] - clock_data["start_time"])
                    / clock_data["num_steps"])
        elif "dt" in clock_data:
            self.dt = clock_data["dt"]
            self.num_steps = (self.end_time - self.start_time) / self.dt
            if not np.isclose(self.num_steps, np.rint(self.num_steps)):
                raise RuntimeError("Simulation interval is not an "
                                   "integer multiple of timestep dt")
            self.num_steps = np.int(np.rint(self.num_steps))

    def advance(self):
        """Increment the time"""
        self.this_step += 1
        self.time = self.start_time + self.dt * self.this_step
        if self.print_time:
            print(f"t = {self.time}")

    def is_running(self):
        """Check if time is less than end time"""
        return self.this_step < self.num_steps


class Grid:
    """Grid class

    Parameters
    ----------
    grid_data : dict
        Dictionary containg parameters needed to defined the grid.
        Currently only 1D grids are defined in turboPy.

        The expected parameters are:

        - ``"N"`` | {``"dr"`` | ``"dx"``} :
            The number of grid points (`int`) | the grid spacing
            (`float`)
        - ``"min"`` | ``"x_min"`` | ``"r_min"`` :
            The coordinate value of the minimum grid point (`float`)
        - ``"max"`` | ``"x_max"`` | ``"r_max"`` :
            The coordinate value of the maximum grid point (`float`)

    Attributes
    ----------
    grid_data : dict
        Dictionary containg parameters needed to defined the grid.
        Currently only 1D grids are defined in turboPy.
    r_min: float, None
        Min of the Grid range.
    r_max : float, None
        Max of the Grid range.
    num_points: int, None
        Number of points on Grid.
    dr : float, None
        Grid spacing.
    r, cell_edges : :class:`numpy.ndarray`
        Array of evenly spaced Grid values.
    cell_centers : float
        Value of the coordinate in the middle of each Grid cell.
    cell_widths : float
        Width of each cell in the Grid.
    r_inv : float
        Inverse of coordinate values at each Grid point,
        1/:class:`Grid.r`.
    """
    def __init__(self, grid_data: dict):
        self.grid_data = grid_data
        self.r_min = None
        self.r_max = None
        self.num_points = None
        self.dr = None
        self.parse_grid_data()

        self.r = (self.r_min + (self.r_max - self.r_min) *
                  self.generate_linear())
        self.cell_edges = self.r
        self.cell_centers = (self.r[1:] + self.r[:-1]) / 2
        self.cell_widths = (self.r[1:] - self.r[:-1])
        with np.errstate(divide='ignore'):
            self.r_inv = 1 / self.r
            self.r_inv[self.r_inv==np.inf] = 0

    def parse_grid_data(self):
        """
        Initializes the grid spacing, range, and number of points on the
        grid from :class:`Grid.grid_data`.

        Raises
        ------
        RuntimeError
            If the range and step size causes a non-integer number of
            grid points.
        """
        self.set_value_from_keys("r_min", {"min", "x_min", "r_min"})
        self.set_value_from_keys("r_max", {"max", "x_max", "r_max"})
        if "N" in self.grid_data:
            self.num_points = self.grid_data["N"]
            self.dr = (self.r_max - self.r_min) / (self.num_points - 1)
        else:
            self.set_value_from_keys("dr", {"dr", "dx"})
            self.num_points = 1 + (self.r_max - self.r_min) / self.dr
            if not (self.num_points % 1 == 0):
                raise (RuntimeError("Invalid grid spacing: "
                                    "configuration does not imply "
                                    "integer number of grid points"))
            self.num_points = np.int(self.num_points)

    def set_value_from_keys(self, var_name, options):
        """
        Initializes a specified attribute to a value provided in
        :class:`Grid.grid_data`.

        Parameters
        ----------
        var_name : str
            Attribute name to be initialized.
        options : set
            Set of keys in :class:`Grid.grid_data` to search for values.

        Raises
        ------
        KeyError
            If none of the keys in `options` are present in
            :class:`Grid.grid_data`.
        """
        for name in options:
            if name in self.grid_data:
                setattr(self, var_name, self.grid_data[name])
                return
        raise (KeyError("Grid configuration for " + var_name
                        + " not found."))

    def generate_field(self, num_components=1, placement_of_points="edge-centered"):
        """Returns squeezed :class:`numpy.ndarray` of zeros with
        dimensions :class:`Grid.num_points` and `num_components`.

        Parameters
        ----------
        num_components : int, defaults to 1
            Number of vector components at each point.
        placement_of_points : str, defaults to "edge-centered"
            Designate position of points on grid
        Returns
        -------
        :class:`numpy.ndarray`
            Squeezed array of zeros.
        """
        number_of_field_points = None
        if placement_of_points == "edge-centered":
            number_of_field_points = self.num_points
        elif placement_of_points == "cell-centered":
            number_of_field_points = self.num_points - 1
        else:
            raise ValueError("Unknown placement option specified")
        return np.squeeze(np.zeros((number_of_field_points, num_components)))

    def generate_linear(self):
        """Returns :class:`numpy.ndarray` with :class:`Grid.num_points`
        evenly spaced in the interval between 0 and 1.

         Returns
         -------
         :class:`numpy.ndarray`
            Evenly spaced array.
         """
        return np.linspace(0, 1, self.num_points)

    def create_interpolator(self, r0):
        """Return a function which linearly interpolates any field on
        this grid, to the point `r0`.

        Parameters
        ----------
        r0 : float
            The requested point on the grid.

        Returns
        -------
        function
            A function which takes a grid quantity `y` and returns the
            interpolated value of `y` at the point `r0`.
        """
        assert (r0 >= self.r_min), "Requested point is not in the grid"
        assert (r0 <= self.r_max), "Requested point is not in the grid"
        i, = np.where((r0 - self.dr < self.r) & (self.r < r0 + self.dr))
        assert (len(i) in [1, 2]), ("Error finding requested point"
                                    "in the grid")
        if len(i) == 1:
            return lambda y: y[i]
        if len(i) == 2:
            # linearly interpolate
            def interpval(yvec):
                """A function which takes a grid quantity `y` and
                returns the interpolated value of `y` at the
                point `r0`.

                Parameters
                ----------
                yvec : :class:`numpy.ndarray`
                    A vector describing a quantity `y` on the grid

                Returns
                -------
                float
                    Value of `y` linearly interpolated to the point `r0`
                """
                rvals = self.r[i]
                y = yvec[i]
                return y[0] + ((r0 - rvals[0]) * (y[1] - y[0])
                               / (rvals[1] - rvals[0]))

            return interpval


class Diagnostic(DynamicFactory):
    """Base diagnostic class.

    Parameters
    ----------
    owner: Simulation
        The Simulation object that owns this object
    input_data: dict
        Dictionary that contains user defined parameters about this
        object such as its name.

    Attributes
    ----------
    _factory_type_name: str
        Type of DynamicFactory child class
    _registry: dict
        Registered derived Diagnostic classes
    owner: Simulation
        The Simulation object that contains this object
    input_data: dict
        Dictionary that contains user defined parameters about this
        object such as its name.
    """

    _factory_type_name = "Diagnostic"
    _registry = {}

    def __init__(self, owner: Simulation, input_data: dict):
        self.owner = owner
        self.input_data = input_data

    def inspect_resource(self, resource: dict):
        """Save references to data from other PhysicsModules

        If your subclass needs the data described by the key, now's
        their chance to save a reference to the data

        Parameters
        ----------
        resource: dict
            A dictionary containing references to data shared by other
            PhysicsModules.
        """
        pass

    def diagnose(self):
        """Perform diagnostic step

        This gets called on every step of the main simulation loop.

        Raises
        ------
        NotImplementedError
            Method or function hasn't been implemented yet. This is an
            abstract base class. Derived classes must implement this
            method in order to be a concrete child class of
            :class:'Diagnostic'.
        """
        raise NotImplementedError

    def initialize(self):
        """Perform any initialization operations

        This gets called once before the main simulation loop.
        """
        pass

    def finalize(self):
        """Perform any finalization operations

        This gets called once after the main simulation loop is
        complete.
        """
        pass



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'turboPy'
# copyright = '2020, Steve Richardson'
html_show_copyright = False
author = 'Steve Richardson'

# The full version, including alpha/beta/rc tags
release = 'v2020.02.21'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx'
]
# Might want to add 'numpydoc', but readthedocs doesn't like it

master_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Options for the napoleon preprocessor
napoleon_google_docstring = False
napoleon_numpy_docstring = True

# Options for intersphinx
intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference/', None),
}

master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'
html_static_path = ['_static']

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    # 'logo': 'logo.png',
    'github_user': 'NRL-Plasma-Physics-Division',
    'github_repo': 'turbopy',
    'description': 'A lightweight framework for computational physics',
    'github_banner': False,
    'github_button': True,
    'travis_button': True,
}

html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html'
    ]
}



"""
Diagnostics module for the turboPy computational physics simulation framework.

Diagnostics can access :class:`PhysicsModule` data. 
They are called every time step, or every N steps.
They can write to file, cache for later, update plots, etc, and they
can halt the simulation if conditions require.
"""
import numpy as np

from .core import Diagnostic, Simulation


class CSVOutputUtility:
    """Comma separated value (CSV) diagnostic output helper class

    Provides routines for writing data to a file in CSV format. This
    class can be used by Diagnostics subclassses to handle output to
    csv format.

    Parameters
    ----------
    filename : str
       File name for CSV data file.
    diagnostic_size : (int, int)
       Size of data set to be written to CSV file. First value is the
       number of time points. Second value is number of spatial points.

    Attributes
    ----------
    filename: str
        File name for CSV data file.
    buffer: :class:`numpy.ndarray`
        Buffer for storing data before it is written to file.
    buffer_index: int
        Position in buffer.
    """

    def __init__(self, filename, diagnostic_size):
        self.filename = filename
        self.buffer = np.zeros(diagnostic_size)
        self.buffer_index = 0

    def append(self, data):
        """Append data to the buffer.

        Parameters
        ----------
        data : :class:`numpy.ndarray`
            1D numpy array of values to be added to the buffer.
        """
        self.buffer[self.buffer_index, :] = data
        self.buffer_index += 1

    def finalize(self):
        """Write the CSV data to file.
        """
        with open(self.filename, 'wb') as f:
            np.savetxt(f, self.buffer, delimiter=",")

class PointDiagnostic(Diagnostic):
    """
    Parameters
    ----------
    owner : Simulation
       Simulation object containing current object.
    input_data : dict
       Dictionary that contains information regarding location, field,
       and output type.

    Attributes
    ----------
    location : str
        Location.
    field_name : str
        Field name.
    output : str
        Output type.
    get_value : function, None
        Function to get value given the field.
    field : str, None
        Field as dictated by resource.
    output_function : function, None
        Function for assigned output method: standard output or csv.
    csv : :class:'numpy.ndarray', None
        numpy.ndarray being written as a csv file.
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.location = input_data["location"]
        self.field_name = input_data["field"]
        self.output = input_data["output_type"]  # "stdout"
        self.get_value = None
        self.field = None
        self.output_function = None
        self.csv = None

    def diagnose(self):
        """
        Run output function given the value of the field.
        """
        self.output_function(self.get_value(self.field))

    def inspect_resource(self, resource):
        """
        Assign attribute field if field_name given in resource.

        Parameters
        ----------
        resource : dict
            Dictionary containing information of field_name to resource.
        """
        if self.field_name in resource:
            self.field = resource[self.field_name]

    def print_diagnose(self, data):
        """
        Prints out data to standard output.

        Parameters
        ----------
        data : :class:`numpy.ndarray`
            1D numpy array of values.
        """
        print(data)

    def initialize(self):
        """
        Initialize output function if provided as csv, and self.csv
        as an instance of the :class: 'CSVOuputUtility' class.
        """
        # set up function to interpolate the field value
        self.get_value = self.owner.grid.create_interpolator(
                                self.location)

        # setup output method
        functions = {"stdout": self.print_diagnose,
                     "csv": self.csv_diagnose,
                     }
        self.output_function = functions[self.input_data["output_type"]]

        if self.input_data["output_type"] == "csv":
            diagnostic_size = (self.owner.clock.num_steps + 1, 1)
            # Use composition to provide csv i/o functionality
            self.csv = CSVOutputUtility(self.input_data["filename"],
                                        diagnostic_size)

    def csv_diagnose(self, data):
        """
        Adds 'data' into csv output buffer.

        Parameters
        ----------
        data : :class:`numpy.ndarray`
            1D numpy array of values.
        """
        self.csv.append(data)

    def finalize(self):
        """
        Write the CSV data to file if CSV is the proper output type.
        """
        self.diagnose()
        if self.input_data["output_type"] == "csv":
            self.csv.finalize()


class FieldDiagnostic(Diagnostic):
    """
    Parameters
    ----------
    owner : Simulation
       Simulation object containing current object.
    input_data : dict
       Dictionary that contains information regarding location, field,
       and output type.

    Attributes
    ----------
    component : str
    field_name : str
        Field.
    output : str
        Output type.
    field : str, None
        Field as dictated by resource.
    dump_interval : SimulationClock, None
        Time interval between two diagnostic runs.
    last_dump : SimulationClock, None
        Time of last diagnostic run.
    diagnose : method
        Run `do_diagnostic` or `check_step` method depending on
        configuration parameters.
    diagnostic_size : (int, int), None
        Size of data set to be written to CSV file. First value is the
        number of time points. Second value is number of spatial points.
    field_was_found : bool
        Boolean representing if field was found in inspect_resource.
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)

        self.component = input_data["component"]
        self.field_name = input_data["field"]
        self.output = input_data["output_type"]  # "stdout"
        self.field = None

        self.dump_interval = None
        self.last_dump = None
        self.diagnose = self.do_diagnostic
        self.diagnostic_size = None

        self.field_was_found = False

    def check_step(self):
        """
        Run diagnostic if dump_interval time has passed since last_dump
        and update last_dump with current time if run.
        """
        if self.owner.clock.time >= self.last_dump + self.dump_interval:
            self.do_diagnostic()
            self.last_dump = self.owner.clock.time

    def do_diagnostic(self):
        """
        Run output_function depending on field.shape.
        """
        if len(self.field.shape) > 1:
            self.output_function(self.field[:, self.component])
        else:
            self.output_function(self.field)

    def inspect_resource(self, resource):
        """
        Assign attribute field if field_name given in resource and
        update boolean if field was found.

        Parameters
        ----------
        resource : dict
            Dictionary containing information of field_name to resource.
        """
        if self.field_name in resource:
            self.field_was_found = True
            self.field = resource[self.field_name]

    def print_diagnose(self, data):
        """
        Print field_name and data onto standard output.

        Parameters
        ----------
        data : :class:`numpy.ndarray`
            1D numpy array of values.
        """
        print(self.field_name, data)

    def initialize(self):
        """
        Initalize diagnostic_size and output function if provided as
        csv, and self.csv as an instance of the
        :class:'CSVOutputUtility' class.
        """
        if not self.field_was_found:
            raise (RuntimeError(f"Diagnostic field {self.field_name}"
                                " was not found"))
        self.diagnostic_size = (self.owner.clock.num_steps + 1,
                                self.field.shape[0])
        if "dump_interval" in self.input_data:
            self.dump_interval = self.input_data["dump_interval"]
            self.diagnose = self.check_step
            self.last_dump = 0
            self.diagnostic_size = (int(np.ceil(
                self.owner.clock.end_time / self.dump_interval) + 1),
                self.field.shape[0])

            # setup output method
        functions = {"stdout": self.print_diagnose,
                     "csv": self.csv_diagnose,
                     }
        self.output_function = functions[self.input_data["output_type"]]
        if self.input_data["output_type"] == "csv":
            self.csv = CSVOutputUtility(self.input_data["filename"],
                                        self.diagnostic_size)

    def csv_diagnose(self, data):
        """
        Adds 'data' into csv output buffer.

        Parameters
        ----------
        data : :class:`numpy.ndarray`
            1D numpy array of values.
        """
        self.csv.append(data)

    def finalize(self):
        """
        Write the CSV data to file if CSV is the proper output type.
        """
        self.do_diagnostic()
        if self.input_data["output_type"] == "csv":
            self.csv.finalize()


class GridDiagnostic(Diagnostic):
    """Diagnostic subclass used to store and save grid data
    into a CSV file

    Parameters
    ----------
    owner : Simulation
        The 'Simulation' object that contains this object
    input_data : dict
        Dictionary containing information about this diagnostic such as
        its name

    Attributes
    ----------
    owner : Simulation
        The 'Simulation' object that contains this object
    input_data : dict
        Dictionary containing information about this diagnostic such as
        its name
    filename : str
        File name for CSV grid file
    """

    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.filename = input_data["filename"]

    def diagnose(self):
        """Grid diagnotic only runs at startup"""
        pass

    def initialize(self):
        """Save grid data into CSV file"""
        with open(self.filename, 'wb') as f:
            np.savetxt(f, self.owner.grid.r, delimiter=",")


class ClockDiagnostic(Diagnostic):
    """Diagnostic subclass used to store and save time data into a CSV
    file using the CSVOutputUtility class.

    Parameters
    ----------
    owner : Simulation
        The 'Simulation' object that contains this object
    input_data : dict
        Dictionary containing information about this diagnostic such as
        its name

    Attributes
    ----------
    owner : Simulation
        The 'Simulation' object that contains this object
    input_data : dict
        Dictionary containing information about this diagnostic such as
        its name
    filename : str
        File name for CSV time file
    csv : :class:'numpy.ndarray'
        Array to store values to be written into a CSV file
    """

    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.filename = input_data["filename"]
        self.csv = None

    def diagnose(self):
        """Append time into the csv buffer."""
        self.csv.append(self.owner.clock.time)

    def initialize(self):
        """Initialize 'self.csv' as an instance of the
        :class:'CSVOuputUtility' class."""
        diagnostic_size = (self.owner.clock.num_steps + 1, 1)
        self.csv = CSVOutputUtility(self.input_data["filename"],
                                    diagnostic_size)

    def finalize(self):
        """Write time into self.csv and saves as a CSV file."""
        self.diagnose()
        self.csv.finalize()


Diagnostic.register("point", PointDiagnostic)
Diagnostic.register("field", FieldDiagnostic)
Diagnostic.register("grid", GridDiagnostic)
Diagnostic.register("clock", ClockDiagnostic)



"""Integration test based on block_on_spring turbopy app

This integration test simply runs the block_on_spring app and compares the output files
to "good" output.
"""
import numpy as np


def test_bos_forwardeuler(bos_run):
    """Tests block_on_spring app with ForwardEuler ComputeTool and compares to
    output files with a "good" output.
    """

    for filename in ['block_p', 'block_x', 'time']:
        ref_data = np.genfromtxt(f'tests/fixtures/block_on_spring/output_forwardeuler/{filename}.csv',
                                 delimiter=',')
        tmp_data = np.genfromtxt(f'tmp/block_on_spring/output_forwardeuler/{filename}.csv',
                                 delimiter=',')
        assert np.allclose(ref_data, tmp_data, rtol=1e-05, atol=1e-08)


def test_bos_backwardeuler(bos_run):
    """Tests block_on_spring app with BackwardEuler ComputeTool and compares to
    output files with a "good" output.
    """

    for filename in ['block_p', 'block_x', 'time']:
        ref_data = np.genfromtxt(f'tests/fixtures/block_on_spring/output_backwardeuler/{filename}.csv',
                                 delimiter=',')
        tmp_data = np.genfromtxt(f'tmp/block_on_spring/output_backwardeuler/{filename}.csv',
                                 delimiter=',')
        assert np.allclose(ref_data, tmp_data, rtol=1e-05, atol=1e-08)


def test_bos_leapfrog(bos_run):
    """Tests block_on_spring app with LeapFrog ComputeTool and compares to
    output files with a "good" output.
    """

    for filename in ['block_p', 'block_x', 'time']:
        ref_data = np.genfromtxt(f'tests/fixtures/block_on_spring/output_leapfrog/{filename}.csv',
                                 delimiter=',')
        tmp_data = np.genfromtxt(f'tmp/block_on_spring/output_leapfrog/{filename}.csv',
                                 delimiter=',')
        assert np.allclose(ref_data, tmp_data, rtol=1e-05, atol=1e-08)



"""Use turboPy to compute the motion of a block on a spring"""
import numpy as np
import pytest
from turbopy import Simulation, PhysicsModule, Diagnostic
from turbopy import CSVOutputUtility, ComputeTool


class BlockOnSpring(PhysicsModule):
    """Use turboPy to compute the motion of a block on a spring"""
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.position = np.zeros((1, 3))
        self.momentum = np.zeros((1, 3))
        self.mass = input_data.get('mass', 1)
        self.spring_constant = input_data.get('spring_constant', 1)
        self.push = owner.find_tool_by_name(input_data["pusher"]).push

    def initialize(self):
        self.position[:] = np.array(self.input_data["x0"])

    def exchange_resources(self):
        self.publish_resource({"Block:position": self.position})
        self.publish_resource({"Block:momentum": self.momentum})

    def update(self):
        self.push(self.position, self.momentum,
                  self.mass, self.spring_constant)


class BlockDiagnostic(Diagnostic):
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.data = None
        self.component = input_data.get("component", 1)
        self.output_function = None
        self.csv = None

    def inspect_resource(self, resource):
        if "Block:" + self.component in resource:
            self.data = resource["Block:" + self.component]

    def diagnose(self):
        self.output_function(self.data[0, :])

    def initialize(self):
        # setup output method
        functions = {"stdout": self.print_diagnose,
                     "csv": self.csv_diagnose,
                     }
        self.output_function = functions[self.input_data["output_type"]]
        if self.input_data["output_type"] == "csv":
            diagnostic_size = (self.owner.clock.num_steps + 1, 3)
            self.csv = CSVOutputUtility(self.input_data["filename"], diagnostic_size)

    def finalize(self):
        self.diagnose()
        if self.input_data["output_type"] == "csv":
            self.csv.finalize()

    def print_diagnose(self, data):
        print(data)

    def csv_diagnose(self, data):
        self.csv.append(data)


class ForwardEuler(ComputeTool):
    """Implementation of the forward Euler algorithm

    y_{n+1} = y_n + h * f(t_n, y_n)
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.dt = None

    def initialize(self):
        self.dt = self.owner.clock.dt

    def push(self, position, momentum, mass, spring_constant):
        p0 = momentum.copy()
        momentum[:] = momentum - self.dt * spring_constant * position
        position[:] = position + self.dt * p0 / mass


class BackwardEuler(ComputeTool):
    """Implementation of the backward Euler algorithm

    y_{n+1} = y_n + h * f(t_{n+1}, y_{n+1})

    Since the position and momentum are separable for this problem, this
    algorithm can be rearranged to give
    alpha = (1 + h^2 * k / m)
    alpha * x_{n+1} = x_n + h * p_n / m
            p_{n+1} = p_n + h * (-k * x_{n+1})
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.dt = None

    def initialize(self):
        self.dt = self.owner.clock.dt

    def push(self, position, momentum, mass, spring_constant):
        factor = 1.0 / (1 + self.dt ** 2 * spring_constant / mass)
        position[:] = (position + self.dt * momentum / mass) * factor
        momentum[:] = momentum - self.dt * spring_constant * position


class Leapfrog(ComputeTool):
    """Implementation of the leapfrog algorithm

    x_{n+1} = x_n + h * fx(t_{n}, p_{n})
    p_{n+1} = p_n + h * fp(t_{n+1}, x_{n+1})
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.dt = None

    def initialize(self):
        self.dt = self.owner.clock.dt

    def push(self, position, momentum, mass, spring_constant):
        position[:] = position + self.dt * momentum / mass
        momentum[:] = momentum - self.dt * spring_constant * position


PhysicsModule.register("BlockOnSpring", BlockOnSpring)
Diagnostic.register("BlockDiagnostic", BlockDiagnostic)
ComputeTool.register("BlockForwardEuler", ForwardEuler)
ComputeTool.register("BackwardEuler", BackwardEuler)
ComputeTool.register("Leapfrog", Leapfrog)


@pytest.fixture
def bos_run():
    # Note: grid isn't used, but "gridless" sims aren't an option yet
    problem_config = {
        "Grid": {"N": 2, "x_min": 0, "x_max": 1},
        "Clock": {"start_time": 0,
                  "end_time": 10,
                  "num_steps": 10},
        "PhysicsModules": {
            "BlockOnSpring": {
                "mass": 1,
                "spring_constant": 1,
                "pusher": "Leapfrog",
                "x0": [0, 1, 0],
                "p0": [0, 0, 0]
            }
        },
        "Tools": {
            "Leapfrog": {},
            "BlockForwardEuler": {},
            "BackwardEuler": {}
        },
        "Diagnostics": {
            # default values come first
            "directory": "tmp/block_on_spring/output_leapfrog/",
            "output_type": "csv",
            "clock": {"filename": "time.csv"},
            "BlockDiagnostic": [
                {'component': 'momentum', 'filename': 'block_p.csv'},
                {'component': 'position', 'filename': 'block_x.csv'}
            ]
        }
    }

    sim = Simulation(problem_config)
    sim.run()

    problem_config["PhysicsModules"]["BlockOnSpring"]["pusher"] = "BlockForwardEuler"
    problem_config["Diagnostics"]["directory"] = "tmp/block_on_spring/output_forwardeuler/"
    sim = Simulation(problem_config)
    sim.run()

    problem_config["PhysicsModules"]["BlockOnSpring"]["pusher"] = "BackwardEuler"
    problem_config["Diagnostics"]["directory"] = "tmp/block_on_spring/output_backwardeuler/"
    sim = Simulation(problem_config)
    sim.run()



"""
Several subclasses of the :class:`turbopy.core.ComputeTool` class for
common scenarios

Included stock subclasses:

- Solver for the 1D radial Poisson's equation
- Helper functions for constructing sparse finite difference matrices
- Charged particle pusher using the Boris method
- Interpolate a function y(x) given y on a grid in x

"""
import numpy as np
import scipy.interpolate as interpolate
from scipy import sparse

from .core import ComputeTool, Simulation


class PoissonSolver1DRadial(ComputeTool):
    """
    Solve 1D radial Poisson's Equation, using finite difference methods

    Parameters
    ----------
    owner : Simulation
        The :class:`turbopy.core.Simulation` object that contains this
        object
    input_data : dict
        There are no custom configuration options for this tool
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
    
    def solve(self, sources):
        """
        Solves Poisson's Equation

        Parameters
        ----------
        sources : :class:`numpy.ndarray`
            Vector containing source terms for the Poisson equation

        Returns
        -------
        :class:`numpy.ndarray`
            Vector containing the finite difference solution
        """
        r = self.owner.grid.r
        dr = np.mean(self.owner.grid.cell_widths)
        I1 = np.cumsum(r * sources * dr)
        integrand = I1 * dr / r
        # linearly extrapolate to r = 0
        i0 = 2 * integrand[1] - integrand[2]
        integrand[0] = i0
        # add constant of integration so derivative = 0 at r = 0
        integrand = integrand - i0
        I2 = np.cumsum(integrand)
        return I2 - I2[-1]


class FiniteDifference(ComputeTool):
    """Helper functions for constructing finite difference matrices

    This class contains functions for constructing finite difference
    approximations to various differential operators. The
    :mod:`scipy.sparse` package from :mod:`scipy` is used since most of
    these are tridiagonal sparse matrices.

    Parameters
    ----------
    owner : Simulation
        The :class:`turbopy.core.Simulation` object that contains this
        object
    input_data : dict
        Dictionary of configuration options.
        The expected parameters are:

        - ``"method"`` | {``"centered"`` | ``"upwind_left"``} :
            Select between centered difference, and left upwind
            difference for the `setup_ddx` member function.
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.dr = self.owner.grid.dr
    
    def setup_ddx(self):
        """Select between centered and upwind finite difference

        Returns
        -------
        function
            Returns a reference to either :meth:`centered_difference` or
            :meth:`upwind_left`, based on the configuration option
            :attr:`input_data["method"]`
        """
        assert (self.input_data["method"] in
                ["centered", "upwind_left"])
        if self.input_data["method"] == "centered":
            return self.centered_difference
        if self.input_data["method"] == "upwind_left":
            return self.upwind_left
    
    def centered_difference(self, y):
        """Centered finite difference estimate for dy/dx

        Parameters
        ----------
        y : :class:`numpy.ndarray`
            Vector of values on the grid

        Returns
        -------
        :class:`numpy.ndarray`
            Estimate of the derivative dy/dx constructed using the
            centered finite difference method
        """
        d = self.owner.grid.generate_field()
        d[1:-1] = (y[2:] - y[:-2]) / (2 * self.dr)
        return d

    def upwind_left(self, y):
        """Left upwind finite difference estimate for dy/dx

        Parameters
        ----------
        y : :class:`numpy.ndarray`
            Vector of values on the grid

        Returns
        -------
        :class:`numpy.ndarray`
            Estimate of the derivative dy/dx constructed using the
            left upwind finite difference method
        """
        d = self.owner.grid.generate_field()
        d[1:] = (y[1:] - y[:-1]) / self.owner.grid.cell_widths
        return d

    def ddx(self):
        """Finite difference matrix for df/dx (centered)

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements the centered finite difference
            approximation to df/dx
        """
        N = self.owner.grid.num_points
        g = 1/(2.0 * self.dr)
        col_below = np.zeros(N) - g
        col_above = np.zeros(N) + g
        D = sparse.dia_matrix(([col_below, col_above], [-1, 1]),
                              shape=(N, N))
        return D

    def radial_curl(self):
        """Finite difference matrix for (rf)'/r = (1/r)(d/dr)(rf)

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a finite difference approximation
            to (rf)'/r = (1/r)(d/dr)(rf)
        """
        N = self.owner.grid.num_points
        g = 1/(2.0 * self.dr)
        col_below = np.zeros(N)
        col_diag = np.zeros(N)
        col_above = np.zeros(N)
        col_below[:-1] = -g * (self.owner.grid.r[:-1]
                               / self.owner.grid.r[1:])
        col_above[1:] = g * (self.owner.grid.r[1:]
                             / self.owner.grid.r[:-1])

        # Set boundary conditions
        # At r=0, use B~linear, and B=0.
        # for col_above, the first element is dropped
        col_above[1] = 2.0 / self.dr
        # At r=Rw, use rB~const?
        # for col_below, the last element is dropped
        col_diag[-1] = 1.0 / self.dr
        col_below[-2] = 2.0 * col_below[-1]
        # set main columns for finite difference derivative
        D = sparse.dia_matrix(([col_below, col_diag, col_above],
                               [-1, 0, 1]), shape=(N, N))
        return D
    
    def del2_radial(self):
        """Finite difference matrix for (1/r)(d/dr)(r (df/dr))

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a finite difference approximation
            to (1/r)(d/dr)(r (df/dr))"""
        N = self.owner.grid.num_points
        g1 = 1/(2.0 * self.dr)
        col_below = -g1 * np.ones(N)
        col_above = g1 * np.ones(N)
        
        col_above[1:] = col_above[1:] / self.owner.grid.r[:-1]
        col_below[:-1] = col_below[:-1] / self.owner.grid.r[1:]
        
        # BC at r=0
        col_above[1] = 0
        
        D1 = sparse.dia_matrix(([col_below, col_above], [-1, 1]),
                               shape=(N, N))
        
        g2 = 1/(self.dr**2)
        col_below = g2 * np.ones(N)
        col_diag = g2 * np.ones(N)
        col_above = g2 * np.ones(N)
        
        # BC at r=0, first row of D
        col_above[1] = 2 * col_above[1]
        D2 = sparse.dia_matrix(([col_below, -2*col_diag, col_above],
                                [-1, 0, 1]), shape=(N, N))
        
        # Need to set boundary conditions!
        D = D1 + D2
        return D
    
    def del2(self):
        """Finite difference matrix for d2/dx2

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a finite difference approximation
            to (d/dx)(df/dx)"""
        N = self.owner.grid.num_points
        
        g2 = 1/(self.dr**2)
        col_below = g2 * np.ones(N)
        col_diag = g2 * np.ones(N)
        col_above = g2 * np.ones(N)
        
        # BC at r=0, first row of D
        col_above[1] = 2 * col_above[1]
        D2 = sparse.dia_matrix(([col_below, -2*col_diag, col_above],
                                [-1, 0, 1]), shape=(N, N))
        return D2

    def ddr(self):
        """Finite difference matrix for (d/dr) f

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a finite difference approximation
            to df/dr
        """
        N = self.owner.grid.num_points
        g1 = 1/(2.0 * self.dr)
        col_below = -g1 * np.ones(N)
        col_above = g1 * np.ones(N)
        # BC at r=0
        col_above[1] = 0
        D1 = sparse.dia_matrix(([col_below, col_above], [-1, 1]),
                               shape=(N, N))
        return D1

    def BC_left_extrap(self):
        """Sparse matrix to extrapolate solution at left boundary

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a boundary condition for the left
            boundary such that the solution at the first two internal
            grid points is extrapolated to the boundary point.
        """
        N = self.owner.grid.num_points
        col_diag = np.ones(N)
        col_above = np.zeros(N)
        col_above2 = np.zeros(N)
        
        # for col_above, the first element is dropped
        col_diag[0] = 0
        col_above[1] = 2
        col_above2[2] = -1

        BC = sparse.dia_matrix(([col_diag, col_above, col_above2],
                                [0, 1, 2]), shape=(N, N))
        return BC

    def BC_left_avg(self):
        """Sparse matrix to set average solution at left boundary

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a boundary condition for the left
            boundary.
        """
        N = self.owner.grid.num_points
        col_diag = np.ones(N)
        col_above = np.zeros(N)
        col_above2 = np.zeros(N)
        
        # for col_above, the first element is dropped
        col_diag[0] = 0
        col_above[1] = 1.5
        col_above2[2] = -0.5

        BC = sparse.dia_matrix(([col_diag, col_above, col_above2],
                                [0, 1, 2]), shape=(N, N))
        return BC        

    def BC_left_quad(self):
        """Sparse matrix for quadratic extrapolation at left boundary

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a boundary condition for the left
            boundary such that the solution at the first two internal
            grid points is extrapolated to the boundary point.
        """
        N = self.owner.grid.num_points
        r = self.owner.grid.r
        col_diag = np.ones(N)
        col_above = np.zeros(N)
        col_above2 = np.zeros(N)
        
        R2 = (r[1]**2 + r[2]**2)/(r[2]**2 - r[1]**2)/2
        # for col_above, the first element is dropped
        col_diag[0] = 0
        col_above[1] = 0.5 + R2
        col_above2[2] = 0.5 - R2

        BC = sparse.dia_matrix(([col_diag, col_above, col_above2],
                                [0, 1, 2]), shape=(N, N))
        return BC
    
    def BC_left_flat(self):
        """Sparse matrix to set Neumann condition at left boundary

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a boundary condition for the left
            boundary such that the derivative of the solution is zero
            at the boundary.
        """
        N = self.owner.grid.num_points
        col_diag = np.ones(N)
        col_above = np.zeros(N)
        # for col_above, the first element is dropped
        col_diag[0] = 0
        col_above[1] = 1

        BC = sparse.dia_matrix(([col_diag, col_above], [0, 1]),
                               shape=(N, N))
        return BC        
    
    def BC_right_extrap(self):
        """Sparse matrix to extrapolate solution at right boundary

        Returns
        -------
        :class:`scipy.sparse.dia_matrix`
            Matrix which implements a boundary condition for the right
            boundary such that the solution at the first two internal
            grid points is extrapolated to the boundary point.
        """
        N = self.owner.grid.num_points
        col_diag = np.ones(N)
        col_below = np.zeros(N)
        col_below2 = np.zeros(N)
        
        # for col_below, the last element is dropped
        col_diag[-1] = 0
        col_below[-2] = 2
        col_below2[-3] = -1

        BC_right = sparse.dia_matrix(([col_below2, col_below, col_diag],
                                      [-2, -1, 0]), shape=(N, N))
        return BC_right


class BorisPush(ComputeTool):
    """
    Calculate charged particle motion in electric and magnetic fields

    This is an implementation of the Boris push algorithm.

    Parameters
    ----------
    owner : Simulation
        The :class:`turbopy.core.Simulation` object that contains this
        object
    input_data : dict
        There are no custom configuration options for this tool

    Attributes
    ----------
    c2 : float
        The speed of light squared
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.c2 = 2.9979e8 ** 2

    def push(self, position, momentum, charge, mass, E, B):
        """
        Update the position and momentum of a charged particle in an
        electromagnetic field

        Parameters
        ----------
        position : :class:`numpy.ndarray`
            The initial position of the particle as a vector
        momentum : :class:`numpy.ndarray`
            The initial momentum of the particle as a vector
        charge : float
            The electric charge of the particle
        mass : float
            The mass of the particle
        E : :class:`numpy.ndarray`
            The value of the electric field at the particle
        B: :class:`numpy.ndarray`
            The value of the magnetic field at the particle
        """
        dt = self.owner.clock.dt

        vminus = momentum + dt * E * charge / 2
        m1 = np.sqrt(mass**2 + np.sum(momentum*momentum, axis=-1)
                     / self.c2)

        t = dt * B * charge / m1[:, np.newaxis] / 2
        s = 2 * t / (1 + np.sum(t*t, axis=-1)[:, np.newaxis])
        
        vprime = vminus + np.cross(vminus, t)
        vplus = vminus + np.cross(vprime, s)
        momentum[:] = vplus + dt * E * charge / 2
        m2 = np.sqrt(mass**2 + np.sum(momentum*momentum, axis=-1)
                     / self.c2)
        position[:] = position + dt * momentum / m2[:, np.newaxis]


class Interpolators(ComputeTool):
    """
    Interpolate a function y(x) given y at grid points in x

    Parameters
    ----------
    owner : Simulation
        The :class:`turbopy.core.Simulation` object that contains this
        object
    input_data : dict
        There are no custom configuration options for this tool
    """
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)

    def interpolate1D(self, x, y, kind='linear'):
        """
        Given two datasets, return an interpolating function

        Parameters
        ----------
        x : list
            List of input values to be interpolated
        y : list
            List of output values to be interpolated
        kind : str
            Order of function being used to relate the two datasets,
            defaults to "linear". Passed as a parameter to
            scipy.interpolate.interpolate.interp1d.

        Returns
        -------
        f : scipy.interpolate.interpolate.interp1d
            Function which interpolates y(x) given grid `x` and
            values `y` on the grid.
        """
        f = interpolate.interp1d(x, y, kind)
        return f


ComputeTool.register("BorisPush", BorisPush)
ComputeTool.register("PoissonSolver1DRadial", PoissonSolver1DRadial)
ComputeTool.register("FiniteDifference", FiniteDifference)
ComputeTool.register("Interpolators", Interpolators)



"""Tests for turbopy/computetools.py"""
import pytest
import numpy as np
from turbopy.computetools import *


def test_init_should_create_object():
    """Method that creates a BorisPush object to test __init__"""
    init_example = BorisPush(Simulation({"type": "BorisPush"}),
                             {"type": "BorisPush"})
    assert isinstance(init_example, BorisPush)
    assert init_example.c2 == 2.9979e8 ** 2


def test_push_with_no_fields():
    """Tests the functionality of the push method, when E=B=0"""
    input_data = {"Clock":
                      {"start_time": 0, "end_time": 100, "dt": 1e-9},
                  "type": "BorisPush",
                  "Grid": {"N": 10, "min": 20, "max": 30},
                  "PhysicsModules": {}}
    owner = Simulation(input_data)
    owner.prepare_simulation()
    push_example = BorisPush(owner, input_data)
    charge = 1.6022e-19
    mass = 1.6726e-27
    E = np.zeros(3)
    B = np.zeros(3)
    x_i = np.array([[0, 0, 0.0]], dtype=np.float)
    v_i = np.array([[0, 0, 3.0e2]], dtype=np.float)
    p_i = mass * v_i
    clock = input_data["Clock"]
    N = 10
    x_final = v_i * N * clock["dt"]
    for i in range(N):
        push_example.push(x_i, p_i, charge, mass, E, B)
    assert np.allclose(x_i, x_final)


def test_push_with_crossed_fields():
    """Tests the functionality of the push method for E, B, not 0"""
    input_data = {
        "Clock": {"start_time": 0, "end_time": 100, "dt": 1e-9},
        "type": "BorisPush",
        "Grid": {"N": 10, "min": 20, "max": 30},
        "PhysicsModules": {}}
    owner = Simulation(input_data)
    owner.prepare_simulation()
    push_example = BorisPush(owner, input_data)
    charge = 1.6022e-19
    mass = 1.6726e-27
    N = 10
    x_i = np.array([[0, 0, 0.0]], dtype=np.float)
    p_i = np.array([[0, 0, 0.0]], dtype=np.float)
    E = np.array([[10, 0, 0]], dtype=np.float)
    B = np.array([[0, 0, 10]], dtype=np.float)
    for i in range(N):
        push_example.push(x_i, p_i, charge, mass, E, B)
    x_final = np.array([[2.20028479e-09, -1.04482737e-08, 0]])
    assert np.allclose(x_i, x_final)
    p_final = np.array([[0.47183691, -1.88168585,  0]]) * mass
    assert np.allclose(p_i / mass, p_final / mass)


@pytest.fixture
def interpolator():
    """Pytest fixture for basic Interpolator class"""
    return Interpolators(Simulation({}), {"type": "Interpolator"})


def test_interpolate1D(interpolator):
    """Tests for turbopy.computetools.Interpolator's interpolate1D method"""
    x = np.arange(0, 10, 1)
    y = np.exp(x)
    xnew = np.arange(0, 1, 0.1)

    f1 = interpolator.interpolate1D(x, y)
    f2 = interpolate.interp1d(x, y)
    assert np.allclose(f1(x), y)
    assert np.allclose(f1(xnew), f2(xnew))

    y = np.asarray([n ** 2 for n in x])
    f1 = interpolator.interpolate1D(x, y, 'quadratic')
    f2 = interpolate.interp1d(x, y, 'quadratic')
    assert np.allclose(f1(x), y)
    assert np.allclose(f1(xnew), f2(xnew))


@pytest.fixture
def fin_diff():
    """Pytest fixture for basic FiniteDifference class with centered method."""
    dic = {"Grid": {"N": 10, "r_min": 0, "r_max": 10},
           "Clock": {"start_time": 0,
                     "end_time": 10,
                     "num_steps": 100},
           "Tools": {},
           "PhysicsModules": {},
           }
    sim = Simulation(dic)
    sim.run()
    return FiniteDifference(sim, {'type': 'FiniteDifference', 'method': 'centered'})


def test_setup_ddx(fin_diff):
    """Tests that `setup_ddx` returns the function specified by `method` in
    `input_data`.
    """
    y = np.arange(0, 10)
    center = fin_diff.setup_ddx()

    assert center == fin_diff.centered_difference
    assert center(y).shape == (10,)
    assert np.allclose(center(y), fin_diff.centered_difference(y))

    fin_diff.input_data['method'] = 'upwind_left'
    upwind = fin_diff.setup_ddx()
    assert upwind == fin_diff.upwind_left
    assert upwind(y).shape == (10,)
    assert np.allclose(upwind(y), fin_diff.upwind_left(y))


def test_centered_difference(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's centered_difference method."""
    dr = fin_diff.owner.grid.dr
    f = fin_diff.owner.grid.generate_field()
    y = np.arange(0, 10)

    assert np.allclose(fin_diff.centered_difference(y),
                       np.append([f[0]], np.append((y[2:] - y[:-2]) / (2 * dr), f[-1])))


def test_upwind_left(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's upwind_left method."""
    cell_widths = fin_diff.owner.grid.cell_widths
    f = fin_diff.owner.grid.generate_field()
    y = np.arange(0, 10)

    assert np.allclose(fin_diff.upwind_left(y), np.append([f[0]], (y[1:] - y[:-1]) / cell_widths))


def test_ddx(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's ddx method."""
    N = fin_diff.owner.grid.num_points
    g = 1 / (2.0 * fin_diff.dr)
    d = fin_diff.ddx()
    assert d.shape == (N, N)
    assert np.allclose(d.toarray(), sparse.dia_matrix(([np.zeros(N) - g, np.zeros(N) + g], [-1, 1]),
                                                      shape=(N, N)).toarray())


def test_radial_curl(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's radial_curl method."""
    with np.errstate(divide='ignore'):
        N = fin_diff.owner.grid.num_points
        dr = fin_diff.owner.grid.dr
        r = fin_diff.owner.grid.r
        g = 1 / (2.0 * dr)
        below = np.append(-g * (r[:-1] / r[1:])[:-1], [0.0, 0.0])
        diag = np.append(np.zeros(N - 1), [1 / dr])
        above = np.append([0.0, 2.0 / dr], g * (r[1:] / r[:-1])[1:])
        d = fin_diff.radial_curl()
        assert d.shape == (N, N)
        for actual, expected in zip(d.data, [below, diag, above]):
            assert np.allclose(actual, expected)


def test_del2_radial(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's del2_radial method."""
    with np.errstate(divide='ignore'):
        N = fin_diff.owner.grid.num_points
        dr = fin_diff.owner.grid.dr
        r = fin_diff.owner.grid.r
        g1 = 1 / (2.0 * dr)
        g2 = 1 / (dr ** 2)
        below = np.append(-g1 / r[1:], [-g1]) + (g2 * np.ones(N))
        above = (np.append([g1, 0], g1 / r[1:-1]) +
                 np.append([g2, g2 * 2], g2 * np.ones(N - 2)))
        diag = -2 * g2 * np.ones(N)
        d = fin_diff.del2_radial()
        d_array = d.toarray()
        assert d.shape == (N, N)
        for ind in range(N - 1):
            assert d_array[ind + 1][ind] == below[ind]
        for ind in range(N):
            assert d_array[ind][ind] == diag[ind]
        for ind in range(N - 1):
            assert d_array[ind][ind + 1] == above[ind + 1]


def test_del2(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's del2 method."""
    N = fin_diff.owner.grid.num_points
    dr = fin_diff.owner.grid.dr
    g = 1 / (dr ** 2)

    below = g * np.ones(N)
    diag = -2 * g * np.ones(N)
    above = g * np.ones(N)
    above[1] *= 2

    d = fin_diff.del2()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [below, diag, above]):
        assert np.allclose(actual, expected)


def test_ddr(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's ddr method."""
    N = fin_diff.owner.grid.num_points
    dr = fin_diff.owner.grid.dr
    g1 = 1 / (2.0 * dr)

    below = -g1 * np.ones(N)
    above = g1 * np.ones(N)
    above[1] = 0

    d = fin_diff.ddr()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [below, above]):
        assert np.allclose(actual, expected)


def test_BC_left_extrap(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's BC_left_extrap method."""
    N = fin_diff.owner.grid.num_points

    diag = np.append(0, np.ones(N-1))
    above = np.append([0.0, 2.0], np.zeros(N-2))
    above2 = np.append([0.0, 0.0, -1.0], np.zeros(N-3))

    d = fin_diff.BC_left_extrap()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [diag, above, above2]):
        assert np.allclose(actual, expected)


def test_BC_left_avg(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's BC_left_avg method."""
    N = fin_diff.owner.grid.num_points

    diag = np.append(0, np.ones(N - 1))
    above = np.append([0.0, 1.5], np.zeros(N - 2))
    above2 = np.append([0.0, 0.0, -0.5], np.zeros(N - 3))

    d = fin_diff.BC_left_avg()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [diag, above, above2]):
        assert np.allclose(actual, expected)


def test_BC_left_quad(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's BC_left_quad method."""
    N = fin_diff.owner.grid.num_points
    r = fin_diff.owner.grid.r
    R = (r[1]**2 + r[2]**2)/(r[2]**2 - r[1]**2)/2

    diag = np.append(0, np.ones(N - 1))
    above = np.append([0.0, 0.5 + R], np.zeros(N - 2))
    above2 = np.append([0.0, 0.0, 0.5 - R], np.zeros(N - 3))

    d = fin_diff.BC_left_quad()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [diag, above, above2]):
        assert np.allclose(actual, expected)


def test_BC_left_flat(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's BC_left_flat method."""
    N = fin_diff.owner.grid.num_points

    diag = np.append(0, np.ones(N - 1))
    above = np.append([0.0, 1], np.zeros(N - 2))

    d = fin_diff.BC_left_flat()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [diag, above]):
        assert np.allclose(actual, expected)


def test_BC_right_extrap(fin_diff):
    """Tests for turbopy.computetools.FiniteDifference's BC_right_extrap method."""
    N = fin_diff.owner.grid.num_points

    below2 = np.append(np.zeros(N - 3), [-1.0, 0.0, 0.0])
    below = np.append(np.zeros(N - 2), [2.0, 0.0])
    diag = np.append(np.ones(N - 1), 0)

    d = fin_diff.BC_right_extrap()
    assert d.shape == (N, N)
    for actual, expected in zip(d.data, [below2, below, diag]):
        assert np.allclose(actual, expected)



"""
Tests for the diagonstics.py file
"""
import pytest
import numpy as np
from turbopy.core import Simulation
from turbopy.diagnostics import FieldDiagnostic


@pytest.fixture(name='simple_field')
def field_fixt():
    """Pytest fixture for FieldDiagnostic class"""
    sim_dic = {"Grid": {"N": 2, "r_min": 0, "r_max": 1},
               "Clock": {"start_time": 0,
                         "end_time": 10,
                         "num_steps": 100},
               "Tools": {"ExampleTool": {}},
               "PhysicsModules": {"ExampleModule": {}}}
    field_dic = {"component": "Component",
                 "field": "Field",
                 "output_type": "csv",
                 "filename": "output.csv",
                 "dump_interval": 1}
    field = FieldDiagnostic(Simulation(sim_dic), field_dic)
    field.owner.read_clock_from_input()
    field.owner.read_grid_from_input()
    return field

#Test methods for FieldDiagnostic class
def test_init_should_create_class_instance_when_called(simple_field):
    """Tests init method in FieldDiagnostic class"""
    assert simple_field.component == "Component"
    assert simple_field.field_name == "Field"
    assert simple_field.output == "csv"
    assert simple_field.field is None
    assert simple_field.dump_interval is None
    assert simple_field.last_dump is None
    assert simple_field.diagnostic_size is None
    assert simple_field.field_was_found is False

def test_check_step_should_update_last_dump_after_dump_interval_has_passed(simple_field):
    """Tests check_step method in FieldDiagnostic class"""
    simple_field.inspect_resource({"Field": np.linspace(0, 1, 2)})
    simple_field.initialize()
    simple_field.owner.clock.time = 1
    simple_field.check_step()
    assert simple_field.last_dump == 1

def test_initialize_should_set_remaining_parameters_when_called(simple_field):
    """Tests initialize method in FieldDiagnostic class"""
    with pytest.raises(RuntimeError):
        assert simple_field.initialize()
    simple_field.inspect_resource({"Field": np.linspace(0, 1, 2)})
    simple_field.initialize()
    assert simple_field.dump_interval == 1
    assert simple_field.last_dump == 0
    assert simple_field.diagnostic_size == (11, 2)
    assert simple_field.output_function == simple_field.csv_diagnose
    assert simple_field.csv.filename == "output.csv"
    assert np.allclose(simple_field.csv.buffer, np.zeros((11, 2)))
    assert simple_field.csv.buffer_index == 0

def test_inspect_resource_should_assign_field_attribute_if_field_name_in_resource(simple_field):
    """Tests inspect_resource method in FieldDiagnostic class"""
    array_to_share = np.linspace(0, 1, 2)
    simple_field.inspect_resource({"Field": array_to_share})
    assert simple_field.field_was_found is True
    assert simple_field.field is array_to_share

def test_csv_diagnose_should_append_data_to_csv_when_called(simple_field):
    """Tests csv_diagnose method in FieldDiagnostic class"""
    simple_field.inspect_resource({"Field": np.linspace(0, 1, 2)})
    simple_field.initialize()
    simple_field.csv.append(simple_field.field)
    assert np.allclose(simple_field.csv.buffer[simple_field.csv.buffer_index-1, :],
                       simple_field.field)
    assert simple_field.csv.buffer_index == 1



import numpy as np
import xarray as xr
# TODO: add tests for plotting
# import matplotlib.pyplot as plt
import pytest

from turbopy import Simulation, PhysicsModule, Diagnostic
from turbopy import CSVOutputUtility, ComputeTool #, FieldDiagnostic
from turbopy import construct_simulation_from_toml


class EMWave(PhysicsModule):
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.c = 2.998e8
        self.E0 = input_data["amplitude"]

        self.E = xr.DataArray(owner.grid.generate_field(),
                              dims="x", coords={"x": owner.grid.r})
        self.E.x.attrs["long_name"] = "Position"
        self.E.x.attrs["units"] = "m"
        self.E.attrs["long_name"] = "Electric Field"
        self.E.attrs["units"] = "V/m"

        self.omega = input_data["omega"]
        self.k = self.omega / self.c
    
    def initialize(self):
        phase = - self.omega * 0 + self.k * (self.owner.grid.r - 0.5)
        self.E[:] = self.E0 * np.cos(2 * np.pi * phase)

    def update(self):
        phase = - self.omega * self.owner.clock.time + self.k * (self.owner.grid.r - 0.5)
        self.E[:] = self.E0 * np.cos(2 * np.pi * phase)
        
    def exchange_resources(self):
        self.publish_resource({"EMField:E": self.E})


class ChargedParticle(PhysicsModule):
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.E = None
        self.x = input_data["position"]
        self.interp_field = owner.grid.create_interpolator(self.x)
        self.position = np.zeros((1, 3))
        self.momentum = np.zeros((1, 3))
        self.eoverm = 1.7588e11
        self.charge = 1.6022e-19
        self.mass = 9.1094e-31
        self.push = owner.find_tool_by_name(input_data["pusher"]).push
        
    def exchange_resources(self):
        self.publish_resource({"ChargedParticle:position": self.position})
        self.publish_resource({"ChargedParticle:momentum": self.momentum})
    
    def inspect_resource(self, resource):
        if "EMField:E" in resource:
            self.E = resource["EMField:E"]    

    def update(self):
        E = np.array([0, self.interp_field(self.E), 0])
        self.push(self.position, self.momentum, self.charge, self.mass, E, B=0)


class ParticleDiagnostic(Diagnostic):
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.data = None
        self.component = input_data["component"]
        self.output_function = None
        
    def inspect_resource(self, resource):
        if "ChargedParticle:" + self.component in resource:
            self.data = resource["ChargedParticle:" + self.component]
    
    def diagnose(self):
        self.output_function(self.data[0, :])

    def initialize(self):
        # setup output method
        functions = {"stdout": self.print_diagnose,
                     "csv": self.csv_diagnose,
                     }
        self.output_function = functions[self.input_data["output_type"]]
        if self.input_data["output_type"] == "csv":
            diagnostic_size = (self.owner.clock.num_steps + 1, 3)
            self.csv = CSVOutputUtility(self.input_data["filename"], diagnostic_size)

    def finalize(self):
        self.diagnose()
        if self.input_data["output_type"] == "csv":
            self.csv.finalize()

    def print_diagnose(self, data):
        print(data)

    def csv_diagnose(self, data):
        self.csv.append(data)


# TODO: add tests for plotting
# class FieldPlottingDiagnostic(FieldDiagnostic):
#     """Extend the FieldDiagnostic to also create plots of the data"""
#     def __init__(self, owner: Simulation, input_data: dict):
#         super().__init__(owner, input_data)
# 
#     def do_diagnostic(self):
#         super().do_diagnostic()
#         plt.clf()
#         self.field.plot()
#         plt.title(f"Time: {self.owner.clock.time:0.3e} s")
#         plt.pause(0.01)
# 
#     def finalize(self):
#         super().finalize()
#         # Call show to keep the plot open
#         plt.show()


class ForwardEuler(ComputeTool):
    def __init__(self, owner: Simulation, input_data: dict):
        super().__init__(owner, input_data)
        self.dt = None
        
    def initialize(self):
        self.dt = self.owner.clock.dt
    
    def push(self, position, momentum, charge, mass, E, B):
        p0 = momentum.copy()
        momentum[:] = momentum + self.dt * E * charge
        position[:] = position + self.dt * p0 / mass


@pytest.fixture
def pif_run():
    PhysicsModule.register("EMWave", EMWave)
    PhysicsModule.register("ChargedParticle", ChargedParticle)
    Diagnostic.register("ParticleDiagnostic", ParticleDiagnostic)
#     Diagnostic.register("FieldPlottingDiagnostic", FieldPlottingDiagnostic)
    ComputeTool.register("ForwardEuler", ForwardEuler)
    input_file = "tests/fixtures/particle_in_field/particle_in_field.toml"
    sim = construct_simulation_from_toml(input_file)
    sim.run()



"""Tests for turbopy/core.py"""
import pytest
import numpy as np
from turbopy.core import *


class ExampleTool(ComputeTool):
    """Example ComputeTool subclass for tests"""


class ExampleModule(PhysicsModule):
    """Example PhysicModule subclass for tests"""
    def update(self):
        pass


# Simulation class test methods
@pytest.fixture(name='simple_sim')
def sim_fixt():
    """Pytest fixture for basic simulation class"""
    dic = {"Grid": {"N": 2, "r_min": 0, "r_max": 1},
           "Clock": {"start_time": 0,
                     "end_time": 10,
                     "num_steps": 100},
           "Tools": {"ExampleTool": {}},
           "PhysicsModules": {"ExampleModule": {}},
           }
    return Simulation(dic)


def test_simulation_init_should_create_class_instance_when_called(simple_sim):
    """Test init method for Simulation class"""
    assert simple_sim.physics_modules == []
    assert simple_sim.compute_tools == []
    assert simple_sim.diagnostics == []
    assert simple_sim.grid is None
    assert simple_sim.clock is None
    assert simple_sim.units is None
    dic = {"Grid": {"N": 2, "r_min": 0, "r_max": 1},
           "Clock": {"start_time": 0,
                     "end_time": 10,
                     "num_steps": 100},
           "Tools": {"ExampleTool": {}},
           "PhysicsModules": {"ExampleModule": {}}
           }
    assert simple_sim.input_data == dic


def test_read_grid_from_input_should_set_grid_attr_when_called(simple_sim):
    """Test read_grid_from_input method in Simulation class"""
    simple_sim.read_grid_from_input()
    assert simple_sim.grid.num_points == 2
    assert simple_sim.grid.r_min == 0
    assert simple_sim.grid.r_max == 1


def test_read_clock_from_input_should_set_clock_attr_when_called(simple_sim):
    """Test read_clock_from_input method in Simulation class"""
    simple_sim.read_clock_from_input()
    assert simple_sim.clock.owner == simple_sim
    assert simple_sim.clock.start_time == 0
    assert simple_sim.clock.time == 0
    assert simple_sim.clock.end_time == 10
    assert simple_sim.clock.this_step == 0
    assert simple_sim.clock.print_time is False
    assert simple_sim.clock.num_steps == 100
    assert simple_sim.clock.dt == 0.1
    dic = {"Grid": {"N": 2, "r_min": 0, "r_max": 1},
           "Clock": {"start_time": 0,
                     "end_time": 10,
                     "dt": 0.2,
                     "print_time": True}}
    other_sim = Simulation(dic)
    other_sim.read_clock_from_input()
    assert other_sim.clock.dt == 0.2
    assert other_sim.clock.num_steps == 50
    assert other_sim.clock.print_time is True


def test_read_tools_from_input_should_set_tools_attr_when_called(simple_sim):
    """Test read_tools_from_input method in Simulation class"""
    ComputeTool.register("ExampleTool", ExampleTool)
    simple_sim.read_tools_from_input()
    assert simple_sim.compute_tools[0].owner == simple_sim
    assert simple_sim.compute_tools[0].input_data == {"type": "ExampleTool"}


def test_fundamental_cycle_should_advance_clock_when_called(simple_sim):
    """Test fundamental_cycle method in Simulation class"""
    simple_sim.read_clock_from_input()
    simple_sim.fundamental_cycle()
    assert simple_sim.clock.this_step == 1
    assert simple_sim.clock.time == 0.1


def test_run_should_run_simulation_while_clock_is_running(simple_sim):
    """Test run method in Simulation class"""
    PhysicsModule.register("ExampleModule", ExampleModule)
    simple_sim.run()
    assert simple_sim.clock.this_step == 100
    assert simple_sim.clock.time == 10


def test_read_modules_from_input_should_set_modules_attr_when_called(simple_sim):
    """Test read_modules_from_input method in Simulation calss"""
    simple_sim.read_modules_from_input()
    assert simple_sim.physics_modules[0].owner == simple_sim
    assert simple_sim.physics_modules[0].input_data == {"name": "ExampleModule"}


#Grid class test methods
@pytest.fixture(name='simple_grid')
def grid_conf():
    """Pytest fixture for grid configuration dictionary"""
    grid = {"N": 8,
            "r_min": 0,
            "r_max": 0.1}
    return Grid(grid)

  
def test_grid_init(simple_grid):
    """Test initialization of the Grid class"""
    assert simple_grid.r_min == 0.0
    assert simple_grid.r_max == 0.1

    
def test_parse_grid_data(simple_grid):
    """Test parse_grid_data method in Grid class"""
    assert simple_grid.num_points == 8
    assert simple_grid.dr == 0.1/7
    # Also test using "dr" to set the grid spacing
    grid_conf2 = {"r_min": 0,
                  "r_max": 0.1,
                  "dr": 0.1/7}
    grid2 = Grid(grid_conf2)
    assert grid2.dr == 0.1/7
    assert grid2.num_points == 8

    
def test_set_value_from_keys(simple_grid):
    """Test set_value_from_keys method in Grid class"""
    assert simple_grid.r_min == 0
    assert simple_grid.r_max == 0.1
    grid_conf1 = {"N": 8,
                  "r_min": 0}
    with pytest.raises(Exception):
        assert Grid(grid_conf1)

        
def test_generate_field(simple_grid):
    """Test generate_field method in Grid class"""
    assert np.allclose(simple_grid.generate_field(), np.zeros(8))
    assert np.allclose(simple_grid.generate_field(3), np.zeros((8, 3)))

    
def test_generate_linear(simple_grid):
    """Test generate_linear method in Grid class"""
    comp = []
    for i in range(simple_grid.num_points):
        comp.append(i/(simple_grid.num_points - 1))
    assert np.allclose(simple_grid.generate_linear(), np.array(comp))

    
def test_create_interpolator(simple_grid):
    """Test create_interpolator method in Grid class"""
    field = simple_grid.generate_linear()
    r_val = 0.05
    interp = simple_grid.create_interpolator(r_val)
    linear_value = r_val / (simple_grid.r_max - simple_grid.r_min)
    assert np.allclose(interp(field), linear_value)


#SimulationClock class test methods
def test_integer_num_steps():
    """Tests for initialization of SimulationClock"""
    clock_config = {'start_time': 0.0,
                    'end_time': 1e-8,
                    'dt': 1e-8 / 10.5,
                    'print_time': True}
    with pytest.raises(RuntimeError):
        SimulationClock(Simulation({}), clock_config)


def test_advance():
    """Tests `advance` method of the SimulationClock class"""
    clock_config = {'start_time': 0.0,
                    'end_time': 1e-8,
                    'num_steps': 20,
                    'print_time': True}
    clock1 = SimulationClock(Simulation({}), clock_config)
    assert clock1.num_steps == (clock1.end_time - clock1.start_time) / clock1.dt
    clock1.advance()
    assert clock1.this_step == 1
    assert clock1.time == clock1.start_time + clock1.dt * clock1.this_step


def test_is_running():
    """Tests `is_running` method of the SimulationClock class"""
    clock_config = {'start_time': 0.0,
                    'end_time': 1e-8,
                    'dt': 1e-8 / 20,
                    'print_time': True}
    clock2 = SimulationClock(Simulation({}), clock_config)
    assert clock2.num_steps == 20
    assert clock2.is_running() and clock2.this_step < clock2.num_steps
    for i in range(clock2.num_steps):
        clock2.advance()
    assert not clock2.is_running()
    



"""Helper functions for constructing Simulations"""
import qtoml as toml

from .core import Simulation

def construct_simulation_from_toml(filename: str) -> Simulation:
    """Construct a Simulation instance from a toml input file

    Parameters
    ----------
    filename : `str`
        The name of the file which contains the input specification, in
        `toml` format.

    Returns
    -------
    simulation_instance : `Simulation`
        An instance of the Simulation class, initialized using the data
        in the input file, which was converted into a python dictionary.
    """
    with open(filename) as f:
        input_data = toml.load(f)

    return Simulation(input_data)


"""Integration test based on particle_in_field turbopy app

This integration test simply runs the particle_in_field app and compares the output files
to "good" output.
"""
import numpy as np


def test_pif(pif_run):
    for filename in ['e_0.5', 'grid', 'particle_p', 'particle_x', 'time']:
        ref_data = np.genfromtxt(f'tests/fixtures/particle_in_field/output/{filename}.csv',
                                 delimiter=',')
        tmp_data = np.genfromtxt(f'tmp/particle_in_field/output/{filename}.csv',
                                 delimiter=',')
        assert (np.allclose(ref_data, tmp_data, rtol=1e-05, atol=1e-08))
